/***************************************************************************
                          c_clustalign.h  -  description
                             -------------------
    begin                : Thu Jun 12 2003
    copyright            : (C) 2003 by Cyril Deremble
    email                : deremble@ihes.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef C_CLUSTALIGN_H
#define C_CLUSTALIGN_H

#include <iostream>
#include "c_analysedseq.h"
#include "c_alignedseqsgroup.h"
#include "c_matrix.h"
#include "max.h"

typedef int score_t;

typedef C_FixedSizeMatrix <short, AMINO_ACID_COUNT, AMINO_ACID_COUNT> aa_subs_matrix_t;

/** class for sequence alignment with the knowledge of the clusters
  *@author Cyril Deremble
  */

class C_ClustAlign
{
	public:
		/** class constructor
		*/
		C_ClustAlign ();
		
		/** align two sequences with the knowledge of the clusters
		*/
		template <class seq1_t, class seq2_t> score_t operator () (seq1_t & Seq1, seq2_t & Seq2) const;
		
		/** get the score of the alignment of two sequences but dont align them;
		*/
		template <class seq1_t, class seq2_t> score_t getAlignScore (const seq1_t & Seq1, const seq2_t & Seq2) const;
		
		/** pointer to the substitution matrix for amino acid not in clusters */
		aa_subs_matrix_t m_SubsMat;
		
		/** pointer to the substitution matrix for amino acid in clusters */
		aa_subs_matrix_t m_ClusterSubsMat;
		
		/** the gap open penalty */
		score_t m_GOP;
		/** the gap extend penalty */
		score_t m_GEP;
		/** the gap open penalty at the end of the sequence */
		score_t m_EndGOP;
		/** the gap extend penalty at the end of the sequence */
		score_t m_EndGEP;
		/** the gap open penalty at inside a cluster */
		score_t m_ClustGOP;
		/** the gap extend penalty at inside a cluster */
		score_t m_ClustGEP;
		
		/** determine wether it uses the overlap or nonoverlap theory */
		bool m_Overlap;
		
		/** setup a global-local alignment */
		inline void setGlobalLocal () {m_EndGOP = m_EndGEP = 0;}
		inline void unsetGlobalLocal () {m_EndGOP = 2 ; m_EndGEP = 1;}
		
	private:
		/** enumeration of the possible directions for the previous cell
		*/
		enum E_Direction {NO_DIR = 0, LEFT_DIR, UP_DIR, LEFTUP_DIR};
		typedef E_Direction dir_t;
		
		/***************************************************************************
		*
		*   strucure that stores the values of the the best alignment to the cell
		*
		*/
		struct S_ScoreCell
		{
			/** the score of the alignment at the cell position
			*/
			score_t Score;
			
			/** the relative position of the previous cell in the matrix (traceback)
			*/
			dir_t PrevCellPos;
			
			/** constructor: initialize the cell
			* direction to the previous cell
			*@param S the score of the cell, default = 0
			*@param Dir the relative position of the previous cell in the matrix, default = no direction
			*/
			S_ScoreCell (score_t S = 0, dir_t Pos = NO_DIR) : Score (S), PrevCellPos (Pos) {}
		};
		
		/* define the matrix of S_ScoreCell */
		typedef C_Matrix <S_ScoreCell> score_matrix_t;
		typedef score_matrix_t::mat_iterator mat_iterator;
		typedef score_matrix_t::mat_const_iterator mat_const_iterator;
		
		mutable score_matrix_t m_Score_mat;
		
		/** used to store the value of (ClustGEP - GEP)
		*/
		mutable score_t m_ClustGEP_GEP;
		
		/** used to store the value of (ClustGOP - GOP)
		*/
		mutable score_t m_ClustGOP_GOP;
		
		void initScoreMat (size_t Size1, size_t Size2) const;
		template <class seq1_t, class seq2_t> void fillScoreMat (const seq1_t & Seq1, const seq2_t & Seq2) const;
		template <class seq1_t, class seq2_t> void traceBack (seq1_t & Seq1, seq2_t & Seq2) const;
		
		inline score_t getGapScore (bool IsGapOpened, float ClusterGapFactor, bool IsSeqEnded) const;
		
		inline float getClusterGapFactor (const C_AnalysedSeq::const_iterator & i, size_t & IsInCluster) const;
		
		template <class T> inline float getClusterGapFactor (const C_GroupIterator <T> & it, size_t & IsInCluster) const;
		
		inline score_t getSubsScore (const C_AnalysedSeq::const_iterator & i, const C_AnalysedSeq::const_iterator & j) const;
		
		template <class T> inline score_t getSubsScore (const C_GroupIterator <T> & it, const C_AnalysedSeq::const_iterator & j) const;
		
		template <class T> inline score_t getSubsScore (const C_GroupIterator <T> & it1, const C_GroupIterator <T> & it2) const;
};


/***************************************************************************
*                                                                         *
*   operator ()
*                                                                         *
***************************************************************************/

template <class seq1_t, class seq2_t> score_t C_ClustAlign::operator () (seq1_t & Seq1, seq2_t & Seq2) const
{
	initScoreMat (Seq1.size () + 1, Seq2.size  () + 1);
	
	fillScoreMat (Seq1, Seq2);
	
	traceBack (Seq1, Seq2);
	
	return m_Score_mat.back ().Score;
}


template <class seq1_t, class seq2_t> score_t C_ClustAlign::getAlignScore (const seq1_t & Seq1, const seq2_t & Seq2) const
{
	initScoreMat (Seq1.size () + 1, Seq2.size  () + 1);
	
	fillScoreMat (Seq1, Seq2);

	return m_Score_mat.back ().Score;
}


/***************************************************************************
*                                                                         *
*   fillScoreMat                                                          *
*                                                                         *
***************************************************************************/

template <class seq1_t, class seq2_t> void C_ClustAlign::fillScoreMat (const seq1_t & Seq1, const seq2_t & Seq2) const
{
	typedef typename seq1_t::const_iterator seq1_iterator;
	typedef typename seq2_t::const_iterator seq2_iterator;
	
	mat_iterator Score_it;
	mat_const_iterator Up_it, Left_it, LeftUp_it;
	
	seq1_iterator i, Last1_it = -- Seq1.end ();
	seq2_iterator j, Last2_it = -- Seq2.end ();
	
	score_t LeftScore, UpScore, LeftUpScore;
	
	m_ClustGOP_GOP = m_ClustGOP - m_GOP;
	m_ClustGEP_GEP = m_ClustGEP - m_GEP;
	
	float ClusterGapFactor1 = 0.0, ClusterGapFactor2 = 0.0;
	bool IsSeq1Ended = false, IsSeq2Ended = false;
	size_t IsSeq1InCluster = 0, IsSeq2InCluster = 0;
	
	for (i = Seq1.begin (), LeftUp_it = m_Score_mat.beginMat (), Up_it = m_Score_mat.beginMat ().colIncr (),
	     Left_it = m_Score_mat.beginMat ().rowIncr (), Score_it = m_Score_mat.beginMat ().diagIncr ();
	     i != Seq1.end (); i ++,
	     Score_it.colIncr (), Up_it.colIncr (), Left_it.colIncr (), LeftUp_it.colIncr ())
	{
		IsSeq1Ended = (i == Last1_it);
		ClusterGapFactor1 = getClusterGapFactor (i, IsSeq1InCluster);
		
		for (j = Seq2.begin (); j != Seq2.end (); j ++,
		     Score_it.colIncr (), Up_it.colIncr (), Left_it.colIncr (), LeftUp_it.colIncr ())
		{
			IsSeq2Ended = (j == Last2_it);
			ClusterGapFactor2 = getClusterGapFactor (j, IsSeq2InCluster);
			
			LeftScore = Left_it->Score - getGapScore (Left_it->PrevCellPos == LEFT_DIR, ClusterGapFactor1, IsSeq1Ended);
			UpScore = Up_it->Score - getGapScore (Up_it->PrevCellPos == UP_DIR, ClusterGapFactor2, IsSeq2Ended);
	  
			LeftUpScore = LeftUp_it->Score + getSubsScore (i, j);
			
			
			Score_it->Score = max (LeftScore, UpScore, LeftUpScore);
			
			Score_it->PrevCellPos = ((Score_it->Score == LeftUpScore) ? LEFTUP_DIR :
				((Score_it->Score == LeftScore) ? LEFT_DIR : UP_DIR));
		}
	}
}


/***************************************************************************
*                                                                         *
*   traceBack                                                             *
*                                                                         *
***************************************************************************/

template <class seq1_t, class seq2_t> void C_ClustAlign::traceBack (seq1_t & Seq1, seq2_t & Seq2) const
{

  typedef typename seq1_t::iterator seq1_iterator;
  typedef typename seq2_t::iterator seq2_iterator;
  

  seq1_iterator i = Seq1.end ();
  seq2_iterator j = Seq2.end ();
  mat_iterator it = m_Score_mat.endMat ();
  -- it;
  
  while (it != m_Score_mat.begin())
    {
      switch (it->PrevCellPos)
	{
	case LEFT_DIR:
	  i = Seq1.insert (i, GAP);
	  -- j;
	  it.colDecr ();
	  break;
	  
	case UP_DIR:
	  j = Seq2.insert (j, GAP);
	  -- i;
	  it.rowDecr ();
	  break;
	  
	case LEFTUP_DIR:
	    it.diagDecr ();
	    -- i;
	    -- j;
	  break;
	  
	case NO_DIR:
	  break;
	}
    }
  
}


/***************************************************************************
*                                                                         *
*   getClusterGapFactor
*                                                                         *
***************************************************************************/

template <class T> float C_ClustAlign::getClusterGapFactor (const C_GroupIterator <T> & it, size_t & IsInCluster) const
{
	typedef typename C_GroupIterator <T>::const_iterator iterator;
	float Tot = 0.0;
	for (iterator i = it.begin (); i != it.end (); i ++)
	{
		if (isFirstInCluster (**i))
			++ IsInCluster;
		else if (isLastInCluster (**i))
			-- IsInCluster;
		++ Tot;
	}
	return (Tot > IsInCluster) ? IsInCluster/Tot : 1.0;
}


/***************************************************************************
*                                                                         *
*   getSubsScore
*                                                                         *
***************************************************************************/

inline score_t C_ClustAlign::getSubsScore (const C_AnalysedSeq::const_iterator & i, const C_AnalysedSeq::const_iterator & j) const
{
	/************************************************************/
	/* determine wether it is the overlap or non-overlap theory */
        return (m_Overlap) ?
		((isInCluster (*i) || isInCluster (*j)) ? m_ClusterSubsMat [*i][*j] : m_SubsMat [*i][*j]) :
		((isInCluster (*i) && isInCluster (*j)) ? m_ClusterSubsMat [*i][*j] : m_SubsMat [*i][*j]);
}

template <class T> score_t C_ClustAlign::getSubsScore (const C_GroupIterator <T> & it, const C_AnalysedSeq::const_iterator & j) const
{
	typedef typename C_GroupIterator <T>::const_iterator iterator;
	float Tot = 0.0;
	score_t Score = 0;
	
	bool Is2InCluster = isInCluster (*j);
	for (iterator i = it.begin (); i != it.end (); i ++)
	{
		/************************************************************/
		/* determine wether it is the overlap or non-overlap theory */
		if (m_Overlap){
			Score += (isInCluster (**i) || Is2InCluster) ? m_ClusterSubsMat [**i][*j] : m_SubsMat [**i][*j];
		}
		else{
			Score += (isInCluster (**i) && Is2InCluster) ? m_ClusterSubsMat [**i][*j] : m_SubsMat [**i][*j];
		}
		++ Tot;
	}
	
	return static_cast <score_t> (Score/Tot);
}


template <class T> score_t C_ClustAlign::getSubsScore (const C_GroupIterator <T> & it1, const C_GroupIterator <T> & it2) const
{
	typedef typename C_GroupIterator <T>::const_iterator iterator;
	float Tot = 0.0;
	score_t Score = 0;
	for (iterator i = it1.begin (); i != it1.end (); i ++)
	{
		bool Is1InCluster = isInCluster (**i);
		for (iterator j = it2.begin (); j != it2.end (); j ++)
		{
		    
                        /************************************************************/
			/* determine wether it is the overlap or non-overlap theory */
			if (m_Overlap){
				Score += (Is1InCluster || isInCluster (**j)) ? m_ClusterSubsMat [**i][**j] : m_SubsMat [**i][**j];
			}
			else{
				Score += (Is1InCluster && isInCluster (**j)) ? m_ClusterSubsMat [**i][**j] : m_SubsMat [**i][**j];
			}
			++ Tot; 
		}
	}
	
	return static_cast <score_t> (Score/Tot);
}


/***************************************************************************
*                                                                         *
*   getGapScore
*                                                                         *
***************************************************************************/

inline score_t C_ClustAlign::getGapScore (bool IsGapOpened, float ClusterGapFactor, bool IsSeqEnded) const
{
	return (IsGapOpened) ?
		((IsSeqEnded) ? m_EndGEP : m_GEP + static_cast <score_t> (ClusterGapFactor*m_ClustGEP_GEP)) :
		((IsSeqEnded) ? m_EndGOP : m_GOP + static_cast <score_t> (ClusterGapFactor*m_ClustGOP_GOP));
}


/***************************************************************************
*                                                                         *
*   getClusterGapFactor
*                                                                         *
***************************************************************************/

inline float C_ClustAlign::getClusterGapFactor (const C_AnalysedSeq::const_iterator & i, size_t & IsInCluster) const
{
	if (isFirstInCluster (*i))
		++ IsInCluster;
	else if (isLastInCluster (*i))
		-- IsInCluster;
	return (IsInCluster > 0) ? 1.0 : 0.0;
}

extern C_ClustAlign clustAlign;


#endif
